#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading

from datetime import datetime
from typing import List

from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor
from fabric.actor.core.manage.management_utils import ManagementUtils
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.orchestrator.core.orchestrator_slice import OrchestratorSlice
from fabric.orchestrator.core.reservation_status_update_thread import ReservationStatusUpdateThread
from fabric.orchestrator.core.slice_defer_thread import SliceDeferThread
from fabric.message_bus.messages.reservation_mng import ReservationMng


class OrchestratorState:
    DELETE_TIMEOUT = 24 * 3600 * 1000

    def __init__(self):
        self.lock = threading.Lock()
        self.used_mac = set()
        self.slices = {}
        self.sdt = None
        self.sut = None
        self.broker = None
        self.logger = None
        self.controller = None

    def set_broker(self, *, broker: str):
        self.broker = broker

    def get_broker(self) -> str:
        return self.broker

    def get_sut(self) -> ReservationStatusUpdateThread:
        return self.sut

    def get_sdt(self) -> SliceDeferThread:
        return self.sdt

    def get_logger(self):
        if self.logger is None:
            from fabric.actor.core.container.globals import GlobalsSingleton
            self.logger = GlobalsSingleton.get().get_logger()
        return self.logger

    def get_management_actor(self) -> IMgmtActor:
        if self.controller is None:
            self.controller = ManagementUtils.get_local_actor()
        return self.controller

    def stop_threads(self):
        if self.sdt is not None:
            self.sdt.stop()

        if self.sut is not None:
            self.sut.stop()

    def start_threads(self):
        self.get_logger().debug("Starting Slice Defer Thread")
        self.sdt = SliceDeferThread()
        self.sdt.start()

        self.get_logger().debug("Starting ReservationStatusUpdateThread")
        self.sut = ReservationStatusUpdateThread()
        self.sut.start()

    def add_slice(self, *, controller_slice: OrchestratorSlice):
        if controller_slice is not None:
            try:
                self.lock.acquire()
                self.slices[controller_slice.get_slice_id()] = controller_slice
            finally:
                self.lock.release()

    def remove_slice(self, *, controller_slice: OrchestratorSlice):
        if controller_slice is not None:
            try:
                self.lock.acquire()
                if controller_slice.get_slice_id() in self.slices:
                    self.slices.pop(controller_slice.get_slice_id())

            finally:
                self.lock.release()

    def get_slice(self, *, slice_id: ID) -> OrchestratorSlice:
        if slice_id is None:
            return None

        try:
            self.lock.acquire()
            return self.slices.get(slice_id, None)
        finally:
            self.lock.release()

    def close_dead_slices(self):
        remove_slices = []
        try:
            self.lock.acquire()
            for slice_id, slice_obj in self.slices.items():
                if slice_obj is None:
                    continue
                try:
                    last_attempt = slice_obj.get_delete_attempt()
                    self.get_logger().debug("Slice {}/{} is_dead_or_closing: {} last_delete_attempt: {}".
                                      format(slice_obj.get_slice_urn(), slice_id,
                                             slice_obj.is_dead_or_closing(), last_attempt))
                    if slice_obj.is_dead_or_closing():
                        if slice_obj.all_failed():
                            if last_attempt is not None:
                                delta_ms = Term.delta(last_attempt, datetime.utcnow())
                                if delta_ms >= self.DELETE_TIMEOUT:
                                    self.get_logger().debug("Deleting all failed slice {}/{} after 24 hours".format(
                                        slice_obj.get_slice_urn(), slice_id))
                                    remove_slices.append(slice_obj)
                            else:
                                slice_obj.mark_delete_attempt()
                        else:
                            self.get_logger().debug("Deleting slice {}/{}".format(slice_obj.get_slice_urn(), slice_id))
                            remove_slices.append(slice_obj)
                except Exception as e:
                    self.get_logger().error("Slice {}/{} owned by {} has encountered a state "
                                      "transition problem at garbage collection".format(slice_obj.get_slice_urn(),
                                                                                        slice_id,
                                                                                        slice_obj.get_user_dn()))
            for slice_obj in remove_slices:
                try:
                    slice_obj.lock()
                    all_res = slice_obj.get_all_reservations(self.get_management_actor())
                    if all_res is not None:
                        for res in all_res:
                            self.release_address_assignment(reservation=res)
                            self.mark_failed_missed_tag(reservation=res)
                    self.remove_slice(controller_slice=slice_obj)
                except Exception as e:
                    self.get_logger().error("Exception occurred in slice cleanup e={}".format(e))
                finally:
                    slice_obj.unlock()
        except Exception as e:
            self.get_logger().error("Unable to close slices due to e={}".format(e))
        finally:
            self.lock.release()

    def release_address_assignment(self, *, reservation: ReservationMng):
        # TODO
        return

    def mark_failed_missed_tag(self, *, reservation: ReservationMng):
        # TODO
        return

    def get_slices(self, *, user_dn: str) -> List[ID]:
        if user_dn is None:
            return None

        result = []
        try:
            self.lock.acquire()
            for slice_id, slice_obj in self.slices.items():
                if slice_obj is not None and slice_obj.get_user_dn() == user_dn:
                    result.append(slice_id)
        finally:
            self.lock.release()

        return result

    def recover(self):
        # TODO
        return

    def recover_slice(self, *, controller: IMgmtActor, slice_id: str, slice_name: str, user_dn: str,
                      ssh_credentials: list):
        # TODO
        return


class OrchestratorStateSingleton:
    __instance = None

    def __init__(self):
        if self.__instance is not None:
            raise Exception("Singleton can't be created twice !")

    def get(self):
        """
        Actually create an instance
        """
        if self.__instance is None:
            self.__instance = OrchestratorState()
        return self.__instance

    get = classmethod(get)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from enum import Enum
from fabric.actor.core.policy.controller_simple_policy import ControllerSimplePolicy
from fabric.actor.core.util.reservation_set import ReservationSet


class TicketReviewSliceState(Enum):
    """
    Redeemable: the default.  No slice reservations have been found that are either Nascent or Failed.
    Nascent: occurs when any reservation is Nascent, i.e. not yet Ticketed.  Will take precedence over Failing.
    Failing: occurs when a slice reservation is found that is Failed.
    """
    Nascent = 1
    Failing = 2
    Redeemable = 3


class ControllerTicketReviewPolicy(ControllerSimplePolicy):
    """
    This implementation of a Service Manager policy is almost identical to the parent ControllerSimplePolicy.
    The only real difference is that it addresses that Tickets should not be redeemed if any reservations are
    currently Failed or Nascent.  This effectively acts as a "gate" between the Controller and AM.
    All reservations must be Ticketed, before any reservations are allowed to be redeemed.
    """
    def __init__(self):
        super().__init__()
        self.pending_redeem = ReservationSet()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['actor']
        del state['clock']
        del state['initialized']
        del state['pending_notify']
        del state['lazy_close']
        del state['pending_redeem']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        self.logger = None
        self.actor = None
        self.clock = None
        self.initialized = False
        self.pending_notify = ReservationSet()
        self.lazy_close = False
        self.pending_redeem = ReservationSet()

    def check_pending(self):
        """
        Check to make sure all reservations are Ticketed (not Failed or Nascent)
        before calling the parent method.

        @throws Exception in case of error
        """
        # add all of our pendingRedeem, so they can be checked
        for reservation in self.pending_redeem.values():
            self.calendar.add_pending(reservation=reservation)

        # get set of reservations that need to be redeemed
        my_pending = self.calendar.get_pending()

        # keep track of status of the slice containing each reservation
        slice_status_map = {}

        # nothing to do!
        if my_pending is None:
            return

        # check the status of the Slice of each reservation
        for reservation in my_pending.values():
            slice_obj = reservation.get_slice()
            slice_id = slice_obj.get_slice_id()

            # only want to do this for 'new' tickets
            if reservation.is_failed() or reservation.is_ticketed():
                # check if we've examined this slice already
                if slice_id not in slice_status_map:
                    # set the default status
                    slice_status_map[slice_id] = TicketReviewSliceState.Redeemable
                    # examine every reservation contained within the slice,
                    # looking for either a Failed or Nascent reservation
                    # we have to look at everything in a slice once, to determine all/any Sites with failures
                    for slice_reservation in slice_obj.get_reservations().values():

                        # If any Reservations that are being redeemed, that means the
                        # slice has already cleared TicketReview.
                        if slice_reservation.is_redeeming() and \
                                slice_status_map[slice_id] == TicketReviewSliceState.Nascent:
                            # There shouldn't be any Nascent reservations, if a reservation is being Redeemed.
                            self.logger.error(
                                "Nascent reservation found while Reservation {} in slice {} is redeeming"
                                .format(slice_reservation.get_reservation_id(), slice_obj.get_name()))

                            # We may have previously found a Failed Reservation,
                            # but if a ticketed reservation is being redeemed,
                            # the failure _should_ be from the AM, not Controller
                            # so it should be ignored by TicketReview
                            slice_status_map[slice_id] = TicketReviewSliceState.Redeemable

                            #  we don't need to look at any other reservations in this slice
                            break

                        # if any tickets are Nascent,
                        # as soon as we remove the Failed reservation,
                        # those Nascent tickets might get redeemed.
                        # we must wait to Close any failed reservations
                        # until all Nascent tickets are either Ticketed or Failed
                        if slice_reservation.is_nascent():
                            self.logger.debug("Found Nascent Reservation {} in slice {} when check_pending for {}"
                                              .format(slice_reservation.get_reservation_id(), slice_obj.get_name(),
                                                      reservation.get_reservation_id()))

                            slice_status_map[slice_id] = TicketReviewSliceState.Nascent
                            # once we have found a Nascent reservation, that is what we treat the entire slice
                            break

                        # track Failed reservations, but need to keep looking for Nascent or Redeemable.
                        if slice_reservation.is_failed():
                            self.logger.debug("Found failed reservation {} in slice {} when check_pending for {}"
                                              .format(slice_reservation.get_reservation_id(), slice_obj.get_name(),
                                                      reservation.get_reservation_id()))
                            slice_status_map[slice_id] = TicketReviewSliceState.Failing

                # take action on the current reservation
                if slice_status_map[slice_id] == TicketReviewSliceState.Failing:
                    if reservation.get_resources() is not None and reservation.get_resources().get_type() is not None:
                        self.logger.info("Closing reservation {} due to failure in slice {}".format(
                            reservation.get_reservation_id(), slice_obj.get_name()))

                        self.actor.close(reservation=reservation)
                        self.calendar.remove_pending(reservation=reservation)
                        self.pending_notify.remove(reservation=reservation)

                elif slice_status_map[slice_id] == TicketReviewSliceState.Nascent:
                    self.logger.info(
                        "Moving reservation {} to pending redeem list due to nascent reservation in slice {}"
                        .format(reservation.get_reservation_id(), slice_obj.get_name()))
                    self.pending_redeem.add(reservation=reservation)
                    self.calendar.remove_pending(reservation=reservation)

        super().check_pending()



import connexion
import six

from fabric.orchestrator.swagger_server.models.success import Success  # noqa: E501
from fabric.orchestrator.swagger_server import util
from fabric.orchestrator.swagger_server.response import resources_controller as rc


def resources_get():  # noqa: E501
    """Retrieve a listing and description of available resources

    Retrieve a listing and description of available resources # noqa: E501


    :rtype: Success
    """
    return rc.resources_get()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from abc import abstractmethod


class ITick:
    """
    ITick defines the interface for objects that can be periodically "ticked".
    A tick is a timer interrupt. Ticks are identified by a
    cycle number. Cycle numbers increase monotonically with every tick. The
    interval between two consecutive ticks is the cycle length. Cycle length is
    fixed and preconfigured at system boot time.

    Implementors of this interface have to take special care to ensure that the
    code in externalTick(long) is efficient, and does not block
    unnecessarily or indefinitely. Blocking inside external tick may have
    significant consequences depending on how ticks are actually delivered.

    It is not guaranteed that cycles passed into two consecutive invocations of
    externalTick(long) will differ with by exactly 1 cycle. It is
    possible for ticks to be delivered not on consecutive cycles. Implementations
    of this interface have to take special care to ensure that they handle
    correctly skipped ticks: e.g., remember the argument of the last invocation
    and perform or operations scheduled to be executed between the last and the
    current cycle.
    """

    @abstractmethod
    def external_tick(self, *, cycle: int):
        """
        Processes a timer interrupt (a tick).

        @param cycle cycle number

        @throws Exception if a critical error occurs while processing the timer
                interrupt. In general, the implementor must catch exceptions
                internally and only pass exceptions up the call stack when
                critical conditions occur.
        """

    @abstractmethod
    def get_name(self) -> str:
        """
        Return the name
        @return name
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod, ABC
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import datetime


class IReservationStatus(ABC):
    """
    IReservationStatus defines a set of predicates that can be used to query the state of a reservation.
    """

    @abstractmethod
    def fail(self, *, message: str, exception: Exception = None):
        """
        Marks an operation failure. Transitions the reservation to the failed state and logs the message as an error.

        Args:
              message: error message
              exception: exception
        """

    @abstractmethod
    def fail_warn(self, *, message: str):
        """
        Marks an operation failure. Transitions the reservation to the failed state and logs the message as an error.

        Args:
              message: error message
        """

    @abstractmethod
    def is_active(self) -> bool:
        """
        Checks if the reservation is active.

        Returns:
            true if the reservation is active
        """

    @abstractmethod
    def is_active_ticketed(self) -> bool:
        """
        Checks if the reservation is activeTicketed.

        Returns:
            true if the reservation is activeTicketed
        """

    @abstractmethod
    def is_closed(self) -> bool:
        """
        Checks if the reservation is closed.

        Returns:
            true if the reservation is closed
        """

    @abstractmethod
    def is_closing(self) -> bool:
        """
        Checks if the reservation is closing.

        Returns:
            true if the reservation is pending closing
        """

    @abstractmethod
    def is_expired(self, *, t: datetime = None) -> bool:
        """
        Checks if the reservation has expired before time t.

        Args:
            t: target date
        Returns:
            true if the reservation has expired before t
        """

    @abstractmethod
    def is_extended(self) -> bool:
        """
        Checks if the reservation has extended at least once.

        Returns:
            true if the reservation has extended at least once
        """

    @abstractmethod
    def is_extending_lease(self) -> bool:
        """
        Checks if the reservation is extending a lease.

        Returns:
            true if the reservation is extending a lease
        """

    @abstractmethod
    def is_extending_ticket(self) -> bool:
        """
        Checks if the reservation is extending a ticket.

        Returns:
            true if the reservation is extending a ticket
        """

    @abstractmethod
    def is_failed(self) -> bool:
        """
        Checks if the reservation has failed.

        Returns:
            true if the reservation has failed
        """

    @abstractmethod
    def is_nascent(self) -> bool:
        """
        Checks if the reservation is nascent.

        Returns:
            true if the reservation pending is nascent
        """

    @abstractmethod
    def is_no_pending(self) -> bool:
        """
        Checks if there is no pending operation.

        Returns:
            true if there is no pending operation
        """

    @abstractmethod
    def is_priming(self) -> bool:
        """
        Checks if the reservation is priming.

        Returns:
            true if the reservation is priming
        """

    @abstractmethod
    def is_redeeming(self) -> bool:
        """
        Checks if the reservation is redeeming.

        Returns:
            true if the reservation is redeeming
        """

    @abstractmethod
    def is_terminal(self) -> bool:
        """
        Checks if the reservation is terminal, e.g., closing, closed, or failed.

        Returns:
            true if the reservation is terminal.
        """

    @abstractmethod
    def is_ticketed(self) -> bool:
        """
        Check if the reservation is ticketed.

        Returns:
            true if the reservation is ticketed.
        """

    @abstractmethod
    def is_ticketing(self) -> bool:
        """
        Checks if the reservation is obtaining a new ticket.

        Returns:
            true if the reservation is obtaining a ticket
        """

    @abstractmethod
    def set_expired(self, *, value: bool):
        """
        Sets the expiration flag.

        Args:
            value: true if the reservation is expired
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker_proxy import IBrokerProxy
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.kernel.resource_set import ResourceSet
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.apis.i_actor import IActor


class IClientReservationFactory:
    """
    Factory for client reservations.
    """
    @staticmethod
    def create(*, rid: ID, resources: ResourceSet = None, term: Term = None, slice_object: ISlice = None,
               broker: IBrokerProxy = None, actor: IActor = None):
        """
        Creates an instance of IClientReservation

        @param rid reservation id
        @param resources resource set
        @param term term
        @param slice_object slice
        @param broker broker
        @param actor actor

        @return an instance of IClientReservation
        """
        raise NotImplementedError("Should have implemented this")

    @staticmethod
    def set_as_source(*, reservation: IClientReservation):
        """
        Updates the reservation to represent a source for a site resource pool.
        @param reservation client reservation
        """
        raise NotImplementedError("Should have implemented this")



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor
from fabric.actor.core.kernel.reservation_states import ReservationStates
from fabric.actor.core.util.id import ID
from fabric.orchestrator.core.exceptions import OrchestratorException
from fabric.orchestrator.core.status_checker import StatusChecker, Status


class ActiveStatusChecker(StatusChecker):
    def __init__(self):
        super().__init__()
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()

    def check_(self, *, controller: IMgmtActor, rid) -> Status:
        if not isinstance(rid, ID):
            return Status.NOTREADY

        try:
            reservation = controller.get_reservation(rid=rid)
            units = controller.get_reservation_units(rid=rid)

            if reservation is None:
                raise OrchestratorException("Unable to obtain reservation information for {}".format(rid))

            if reservation.get_state() == ReservationStates.Active:
                if units is None or len(units) == 0:
                    return Status.NOTREADY
            elif reservation.get_state() == ReservationStates.Closed:
                return Status.OK
            elif reservation.get_state() == ReservationStates.Failed:
                return Status.NOTOK

        except Exception as e:
            self.logger.error("Exception occurred e: {}".format(e))
        return Status.NOTREADY



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from typing import List

from fabric.actor.core.util.id import ID


class IStatusUpdateCallback:
    """
    Callbacks on reservation status updates must comply with this interface
    """
    def success(self, *, ok: List[ID], act_on: List[ID]):
        """
        All reservations in indicated group have gone to Active or have OK modify status
        @param ok - reservations that transitioned to Active or OK modify status
        @param act_on - reservations that need to be acted on
        @throws Exception in case of error
        """
        raise NotImplementedError

    def failure(self, *, failed: List[ID], ok: List[ID], act_on: List[ID]):
        """
        Some reservations may have gone into Failed or not OK modify status, so provide an action for those
        @param failed - those reservations that failed or went to not OK
        @param ok - reservations in the same group that went Active or OK
        @param act_on - reservations to be acted on
        @throws Exception in case of error
        """
        raise NotImplementedError



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fim.graph.abc_property_graph import ABCPropertyGraph

from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.apis.i_server_public import IServerPublic

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_client_callback_proxy import IClientCallbackProxy
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.util.client import Client
    from fabric.actor.core.util.id import ID
    from fabric.actor.security.auth_token import AuthToken


class IServerActor(IActor, IServerPublic):
    """
    IServerActor defines the common functionality for actors
    acting as servers for other actors (brokers and site authorities).
    """
    @abstractmethod
    def donate_reservation(self, *, reservation: IClientReservation):
        """
        Accepts ticketed resources to be used for allocation of client
        requests.

        @params reservation: reservation representing resources to be used for
               allocation

        @raises Exception in case of error
        """

    @abstractmethod
    def register_client_slice(self, *, slice_obj: ISlice):
        """
        Registers a new client slice.
        @params slice_obj: client slice
        @raises Exception in case of error
        """

    @abstractmethod
    def register_client(self, *, client: Client):
        """
        Registers the specified client.
        @param client client to register
        @throws Exception in case of error
        """

    @abstractmethod
    def unregister_client(self, *, guid: ID):
        """
        Unregisters the specified client.
        @params guid : client guid
        @raises Exception in case of error
        """

    @abstractmethod
    def get_client(self, *, guid: ID) -> Client:
        """
        Get a client specified by GUID
        @params guid: guid
        @returns client: specified by GUID
        @raises Exception in case of error
        """

    @abstractmethod
    def claim_delegation(self, *, delegation: IDelegation, callback: IClientCallbackProxy, caller: AuthToken,
                         id_token: str = None):
        """
        Processes an incoming claim request.
        @params delegation: delegation
        @params callback : callback
        @params caller: caller
        @param id_token id token
        @raises Exception in case of error
        """

    @abstractmethod
    def reclaim_delegation(self, *, delegation: IDelegation, callback: IClientCallbackProxy, caller: AuthToken,
                           id_token: str = None):
        """
        Processes an incoming claim request.
        @params delegation: delegation
        @params callback : callback
        @params caller: caller
        @param id_token id token
        @raises Exception in case of error
        """

    @abstractmethod
    def ticket(self, *, reservation: IReservation, callback: IClientCallbackProxy, caller: AuthToken):
        """
        Processes an incoming ticket request.
        @params reservation: reservation
        @params callback : callback
        @params caller: caller
        @raises Exception in case of error
        """

    @abstractmethod
    def extend_ticket(self, *, reservation: IReservation, caller: AuthToken):
        """
        Processes an incoming extend_ticket request.
        @params reservation: reservation
        @params caller: caller
        @raises Exception in case of error
        """

    @abstractmethod
    def relinquish(self, *, reservation: IReservation, caller: AuthToken):
        """
        Processes an incoming relinquish request.
        @params reservation: reservation
        @params caller: caller
        @raises Exception in case of error
        """

    @abstractmethod
    def advertise(self, *, delegation: ABCPropertyGraph, client: AuthToken) -> ID:
        """
        Exports the resources described by the delegation to the client.
        @param delegation delegation describing resources to export
        @param client identity of the client resources will be exported to
        @raises Exception in case of error
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_callback_proxy import ICallbackProxy
from fabric.actor.core.apis.i_rpc_request_state import IRPCRequestState
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ProxyException
from fabric.actor.core.core.rpc_request_state import RPCRequestState
from fabric.actor.core.kernel.incoming_failed_rpc import IncomingFailedRPC
from fabric.actor.core.kernel.incoming_query_rpc import IncomingQueryRPC
from fabric.actor.core.kernel.incoming_reservation_rpc import IncomingReservationRPC
from fabric.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
from fabric.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric.actor.core.proxies.proxy import Proxy
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.util.rpc_exception import RPCError
from fabric.actor.security.auth_token import AuthToken


class LocalProxy(Proxy, ICallbackProxy):
    class LocalProxyRequestState(RPCRequestState):
        def __init__(self):
            super().__init__()
            self.reservation = None
            self.update_data = None
            self.callback = None
            self.query = None
            self.request_id = None
            self.failed_reservation_id = None
            self.failed_request_type = None
            self.error_detail = None
            self.delegation = None

    def __init__(self, *, actor: IActor):
        super().__init__(auth=actor.get_identity())
        self.logger = actor.get_logger()
        self.proxy_type = Constants.protocol_local

    def execute(self, *, request):
        try:
            incoming = None
            if request.get_type() == RPCRequestType.Query:
                incoming = IncomingQueryRPC(request_type=RPCRequestType.Query, message_id=request.get_message_id(),
                                            query=request.query, caller=request.get_caller(),
                                            id_token=request.get_id_token(), callback=request.callback)

            elif request.get_type() == RPCRequestType.QueryResult:
                incoming = IncomingQueryRPC(request_type=RPCRequestType.QueryResult,
                                            message_id=request.get_message_id(), query=request.query,
                                            id_token=request.get_id_token(),
                                            caller=request.get_caller(), request_id=request.request_id)

            elif request.get_type() == RPCRequestType.Ticket or request.get_type() == RPCRequestType.Redeem or \
                    request.get_type() == RPCRequestType.ExtendTicket or \
                    request.get_type() == RPCRequestType.ExtendLease or \
                    request.get_type() == RPCRequestType.Close or request.get_type() == RPCRequestType.Relinquish:
                incoming = IncomingReservationRPC(message_id=request.get_message_id(), request_type=request.get_type(),
                                                  reservation=request.reservation,
                                                  callback=request.callback, caller=request.get_caller())

            elif request.get_type() == RPCRequestType.UpdateTicket or request.get_type() == RPCRequestType.UpdateLease:
                incoming = IncomingReservationRPC(message_id=request.get_message_id(), request_type=request.get_type(),
                                                  reservation=request.reservation,
                                                  update_data=request.update_data, caller=request.get_caller())

            elif request.get_type() == RPCRequestType.FailedRPC:
                incoming = IncomingFailedRPC(message_id=request.get_message_id(),
                                             failed_request_type=request.failed_request_type,
                                             request_id=request.request_id,
                                             failed_reservation_id=request.failed_reservation_id,
                                             error_details=request.error_detail, caller=request.get_caller())
            else:
                raise ProxyException("Unsupported RPC type: {}".format(request.get_type()))
            RPCManagerSingleton.get().dispatch_incoming(actor=self.get_actor(), rpc=incoming)

        except Exception as e:
            raise ProxyException("Error while processing RPC request{} {}".format(RPCError.InvalidRequest, e))

    def prepare_query(self, *, callback: ICallbackProxy, query: dict, caller: AuthToken, id_token: str):
        state = self.LocalProxyRequestState()
        state.id_token = id_token
        state.query = query
        state.callback = callback
        return state

    def prepare_query_result(self, *, request_id: str, response, caller: AuthToken) -> IRPCRequestState:
        state = self.LocalProxyRequestState()
        state.query = response
        state.request_id = request_id
        return state

    def prepare_failed_request(self, *, request_id: str, failed_request_type,
                               failed_reservation_id, error: str, caller: AuthToken) -> IRPCRequestState:
        state = self.LocalProxyRequestState()
        state.request_id = request_id
        state.failed_request_type = failed_request_type
        state.failed_reservation_id = failed_reservation_id
        state.error_detail = error
        return state

    def get_actor(self) -> IActor:
        result = ActorRegistrySingleton.get().get_actor(self.get_name())
        if result is None:
            raise ProxyException("Actor does not exist.")
        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_authority_public import IAuthorityPublic
from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.apis.i_server_actor import IServerActor

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
    from fabric.actor.core.apis.i_controller_callback_proxy import IControllerCallbackProxy
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.kernel.resource_set import ResourceSet
    from fabric.actor.security.auth_token import AuthToken


class IAuthority(IServerActor, IAuthorityPublic):
    """
    IAuthority defines the interface for an actor acting in the authority role.
    """
    @abstractmethod
    def available(self, *, resources: ResourceSet):
        """
        Informs the actor that the following resources are available for
        allocation.

        @param resources resources

        @raises Exception in case of error
        """

    @abstractmethod
    def donate(self, *, resources: ResourceSet):
        """
        Accepts concrete resources to be used for allocation of client
        requests.

        @param resources resource set representing resources to be used for
               allocation

        @raises Exception in case of error
        """

    @abstractmethod
    def donate_delegation(self, *, delegation: IDelegation):
        """
        Accepts delegations to be merged to Model.

        @param delegation delegation

        @raises Exception in case of error
        """

    @abstractmethod
    def eject(self, *, resources: ResourceSet):
        """
        Ejects the specified resources from the inventory.

        @param resources resources

        @raises Exception in case of error
        """

    @abstractmethod
    def extend_lease(self, *, reservation: IAuthorityReservation, caller: AuthToken = None):
        """
        Processes an extend lease request for the reservation.

        @param reservation reservation representing a request for a lease
               extension
        @param caller : caller

        @raises Exception in case of error
        """

    @abstractmethod
    def modify_lease(self, *, reservation: IAuthorityReservation, caller: AuthToken):
        """
        Processes an modify lease request for the reservation.

        @param reservation :  reservation representing a request for a lease
               modification
        @param caller : caller

        @raises Exception in case of error
        """

    @abstractmethod
    def freed(self, *, resources: ResourceSet):
        """
        Informs the actor that the given resources are no longer in use
        and can be considered as free, regardless of the state of the
        individual units.

        @param resources resource set representing freed resources

        @raises Exception in case of error
        """

    @abstractmethod
    def redeem(self, *, reservation: IReservation, callback: IControllerCallbackProxy = None, caller: AuthToken = None):
        """
        Processes a redeem request for the reservation.

        @param reservation reservation representing a request for a new lease
        @param callback callback
        @param caller caller
        @raises Exception in case of error
        """

    @abstractmethod
    def unavailable(self, *, resources: ResourceSet) -> int:
        """
        Informs the actor that previously donated resources are no
        longer available for allocation.

        @param resources resources

        @return number of unavailable units

        @raises Exception in case of error
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING, List

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import PolicyException
from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
from fabric.actor.core.core.unit_set import UnitSet
from fabric.actor.core.policy.resource_control import ResourceControl
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.ipv4_set import IPv4Set
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.core.unit import Unit, UnitState

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.util.resource_type import ResourceType
    from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
    from fabric.actor.core.apis.i_reservation import IReservation


class Vmm:
    def __init__(self, *, host: Unit, capacity: int):
        if host is None:
            raise PolicyException(Constants.not_specified_prefix.format("host"))
        if capacity < 1:
            raise PolicyException("capacity must be at least 1")
        self._host = host
        self.hosted = set()
        self.capacity = capacity

    def __str__(self):
        return "{} {}".format(self._host, self.capacity)

    def __eq__(self, other):
        if not isinstance(other, Vmm):
            return False
        return self._host.get_id() == other._host.get_id()

    def __hash__(self):
        return self._host.get_id().__hash__()

    def release(self, *, vm: Unit):
        if vm is None:
            raise PolicyException(Constants.not_specified_prefix.format("vm"))

        if vm not in self.hosted:
            raise PolicyException("the specified node is not hosted on this vmm")

        self.hosted.remove(vm)

    def host(self, *, vm: Unit):
        if vm is None:
            raise PolicyException(Constants.not_specified_prefix.format("vm"))

        if vm in self.hosted:
            raise PolicyException("the specified node is already hosted on this vmm")

        self.hosted.add(vm)

    def get_hosted_count(self) -> int:
        return len(self.hosted)

    def get_capacity(self) -> int:
        return self.capacity

    def get_available(self) -> int:
        return self.capacity - len(self.hosted)

    def get_host(self) -> Unit:
        return self._host

    def get_hosted_vms(self) -> set:
        return self.hosted.copy()


class VmmPool:
    def __init__(self, *, rtype: ResourceType, properties: dict):
        self.rtype = rtype
        self.properties = properties
        self.vmms = {}
        self.memory = 0
        self.cpu = 0
        self.bandwidth = 0
        self.disk = 0
        self.capacity = 0

    def donate(self, *, vm: Vmm):
        if vm is None:
            raise PolicyException(Constants.not_specified_prefix.format("vm"))

        if vm.get_host().get_actor_id() in self.vmms:
            raise PolicyException("the specified vm already in the pool")

        self.vmms[vm.get_host().get_id()] = vm

    def get_vmm_set(self) -> List[Vmm]:
        result = []
        for v in self.vmms.values():
            result.append(v)
        return result

    def get_vmm(self, *, uid: ID) -> Vmm:
        if uid in self.vmms:
            return self.vmms[uid]
        return None

    def get_vmms_count(self) -> int:
        return len(self.vmms)

    def get_properties(self):
        return self.properties

    def get_memory(self) -> int:
        return self.memory

    def set_memory(self, *, memory: int):
        self.memory = memory

    def get_cpu(self) -> int:
        return self.cpu

    def set_cpu(self, *, cpu: int):
        self.cpu = cpu

    def get_bandwidth(self) -> int:
        return self.bandwidth

    def set_bandwidth(self, *, bandwidth: int):
        self.bandwidth = bandwidth

    def get_disk(self) -> int:
        return self.disk

    def set_disk(self, *, disk: int):
        self.disk = disk

    def get_capacity(self) -> int:
        return self.capacity

    def set_capacity(self, *, capacity: int):
        self.capacity = capacity

    def get_type(self) -> ResourceType:
        return self.rtype


class VMControl(ResourceControl):
    PropertyCapacity = "capacity"
    PropertyIPList = "ip.list"
    PropertyIPSubnet = "ip.subnet"
    PropertyIPGateway = "ip.gateway"
    PropertyDataSubnet = "data.subnet"

    def __init__(self):
        super().__init__()
        self.inventory = {}
        self.ipset = IPv4Set()
        self.subnet = None
        self.gateway = None
        self.use_ip_set = False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['authority']
        del state['logger']
        del state['initialized']

        del state['inventory']
        del state['ipset']
        del state['subnet']
        del state['gateway']
        del state['use_ip_set']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        self.authority = None
        self.logger = None
        self.initialized = False

        self.inventory = {}
        self.ipset = IPv4Set()
        self.subnet = None
        self.gateway = None
        self.use_ip_set = False

    def donate_reservation(self, *, reservation: IClientReservation):
        return

    def donate(self, *, resource_set: ResourceSet):
        rtype = resource_set.get_type()
        resource = resource_set.get_resource_properties()
        local = resource_set.get_local_properties()

        pool = self.inventory.get(rtype)
        if pool is None:
            pool = VmmPool(rtype=rtype, properties=resource)
            rd = ResourcePoolDescriptor()
            rd.reset(properties=resource)
            memory = int(rd.get_attribute(key=Constants.resource_memory).get_value())
            capacity = 0
            if self.PropertyCapacity in local:
                capacity = int(local[self.PropertyCapacity])
            if self.PropertyIPSubnet in local:
                self.subnet = local[self.PropertyIPSubnet]
            if self.PropertyIPGateway in local:
                self.gateway = local[self.PropertyIPGateway]

            if self.PropertyIPList in local:
                temp = local[self.PropertyIPList]
                self.ipset.add(ip_list=temp)
                self.use_ip_set = True

            pool.set_memory(memory=memory)
            pool.set_capacity(capacity=capacity)
            self.inventory[rtype] = pool

        unit_set = resource_set.get_resources()
        for unit in unit_set.get_set().values():
            vmm = Vmm(host=unit, capacity=pool.get_capacity())
            pool.donate(vm=vmm)

    def assign(self, *, reservation: IAuthorityReservation) -> ResourceSet:
        reservation.set_send_with_deficit(value=True)

        if len(self.inventory) == 0:
            raise PolicyException("no inventory")

        requested = reservation.get_requested_resources()
        request_properties = requested.get_request_properties()
        rtype = requested.get_type()
        current = reservation.get_resources()
        ticket = requested.get_resources()

        gained = None
        lost = None

        if current is None:
            pool = self.inventory.get(rtype)
            if pool is None:
                raise PolicyException("no resources of the specified pool")

            needed = ticket.get_units()
            gained = self.get_vms(pool=pool, needed=needed)

            if gained is None or gained.get_units() == 0:
                self.logger.warning("Could not allocate any units for r: {}".format(reservation.get_reservation_id()))
                return None
        else:
            rtype = current.get_type()
            pool = self.inventory.get(rtype)
            current_units = current.get_units()
            difference = ticket.get_units() - current_units

            if difference > 0:
                gained = self.get_vms(pool=pool, needed=difference)
            elif difference < 0:
                unit_set = current.get_resources()
                victims = request_properties[Constants.config_victims]
                to_take = unit_set.select_extract(count=-difference, victims=victims)
                lost = UnitSet(plugin=self.authority.get_plugin(), units=to_take)

        return ResourceSet(gained=gained, lost=lost, rtype=rtype)

    def get_vms(self, *, pool: VmmPool, needed: int) -> UnitSet:
        uset = UnitSet(plugin=self.authority.get_plugin())
        vmms = pool.get_vmm_set()
        allocated = 0

        for vmm in vmms:
            available = vmm.get_available()
            if self.use_ip_set:
                available = min(available, self.ipset.get_free_count())

            if available > 0:
                to_allocate = min(available, needed - allocated)
                for i in range(to_allocate):
                    vm = Unit(id=ID())
                    vm.set_resource_type(rtype=pool.get_type())
                    vm.set_parent_id(parent_id=vmm.get_host().get_id())
                    vm.set_property(name=Constants.unit_parent_host_name,
                                    value=vmm.get_host().get_property(name=Constants.unit_host_name))
                    vm.set_property(name=Constants.unit_control,
                                    value=vmm.get_host().get_property(name=Constants.unit_control))
                    vm.set_property(name=Constants.unit_memory,
                                    value=str(pool.get_memory()))
                    if self.use_ip_set:
                        vm.set_property(name=Constants.unit_management_ip,
                                        value=self.ipset.allocate())

                    if self.subnet is not None:
                        vm.set_property(name=Constants.unit_manage_subnet, value=self.subnet)

                    if self.gateway is not None:
                        vm.set_property(name=Constants.unit_manage_gateway, value=self.gateway)

                    vmm.host(vm=vm)
                    uset.add_unit(u=vm)
                allocated += to_allocate
            if allocated == needed or self.ipset.get_free_count() == 0:
                break

        return uset

    def free(self, *, uset: dict):
        if uset is not None:
            for u in uset.values():
                try:
                    rtype = u.get_resource_type()
                    pool = self.inventory.get(rtype)
                    host = u.get_parent_id()
                    vmm = pool.get_vmm(uid=host)
                    vmm.release(vm=u)
                    if self.use_ip_set:
                        self.ipset.free(ip=u.get_property(Constants.unit_management_ip))
                    self.logger.debug("Released unit: {}".format(u))
                except Exception as e:
                    self.logger.error("Failed to release unit: {} exception:{}".format(u, e))

    def revisit(self, *, reservation: IReservation):
        unit_set = reservation.get_resources().get_resources()
        for u in unit_set.get_set().values:
            try:
                if u.get_state() == UnitState.DEFAULT or \
                        u.get_state() == UnitState.FAILED or \
                        u.get_state() == UnitState.CLOSING or \
                        u.get_state() == UnitState.PRIMING or \
                        u.get_state() == UnitState.ACTIVE or \
                        u.get_state() == UnitState.MODIFYING:
                    rtype = u.get_resource_type()
                    pool = self.inventory.get(rtype)
                    uid = u.get_parent_id()
                    vmm = pool.get_vmm(uid=uid)
                    vmm.host(vm=u)
                    self.logger.debug("VMControl.revisit(); recovering management IP {}".format(
                        u.get_property(name=Constants.unit_management_ip)))
                    self.ipset.reserve(ip=u.get_property(name=Constants.unit_management_ip))
            except Exception as e:
                self.fail(u=u, message="revisit with vmcontrol", e=e)

    def recovery_starting(self):
        self.logger.info("Beginning VMControl recovery")

    def recovery_ended(self):
        self.logger.info("Completing VMControl recovery")
        self.logger.debug("Restored VMControl with subnet {} gateway {} ipset {} and inventory {}"
                          .format(self.subnet, self.gateway, self.ipset, self.inventory))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import bisect

from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.util.reservation_set import ReservationSet
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.core.util.utils import binary_search


class ReservationWrapper:
    """
    Internal class to represent a reservation.
    """
    def __init__(self, *, reservation: IReservation, start: int, end: int):
        self.start = start
        self.end = end
        self.reservation = reservation

    def __eq__(self, other):
        if not isinstance(other, ReservationWrapper):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return (self.start == other.start and
                self.end == other.end and
                self.reservation.get_reservation_id() == other.reservation.get_reservation_id())

    def __lt__(self, other):
        if not isinstance(other, ReservationWrapper):
            # don't attempt to compare against unrelated types
            return NotImplemented

        result = 0
        if self.end == other.end:
            if self.reservation is not None and other.reservation is not None:
                result = self.reservation.get_reservation_id() < other.reservation.get_reservation_id()

            return result
        return self.end < other.end

    def __str__(self):
        return "{}:{}:{}:{}".format(self.end, self.start, self.end, self.reservation.get_reservation_id())


class ReservationHoldings:
    """
    This class maintains a collection of reservations. Each reservation is
    associated with a validity interval. The class allows to answer intersection
    queries: what reservations are valid at a given time instance.

    As time goes by, the class can be purged from irrelevant reservation records
    by invoking tick(). Purging is strongly recommended as it reduces the cost of intersection queries.

    An attempt has been made to optimize the cost of using this data structure.
    Inserts are O(log(n)). Queries, however, may take between O(log(n)) and O(n).
    """
    def __init__(self):
        # List of reservation wrappers sorted by increasing end time.
        self.list = []
        # All reservations stored in this collection.
        self.reservation_set = ReservationSet()
        # Map of reservations to ReservationWrappers. Needed when removing a reservation.
        self.map = {}

    def add_reservation(self, *, reservation: IReservation, start: int, end: int):
        """
        Adds a reservation to the collection for the specified period of time.
        The interval is closed on both sides.
        @params reservation : reservation to add
        @params start : start time
        @params end : end time
        """
        # If this is an extended reservation, we may already have it in the
        # list (with potentially different start and end times). Remove the
        # previous entry if this is the case.
        my_start = start
        entry = None
        if reservation.get_reservation_id() in self.map:
            entry = self.map[reservation.get_reservation_id()]
            if entry is not None:
                assert (start - entry.end) <= 1
                my_start = entry.start
                self.remove_reservation(reservation=reservation)

        entry = ReservationWrapper(reservation=reservation, start=my_start, end=end)
        self.add_to_list(entry=entry)
        self.reservation_set.add(reservation=reservation)
        self.map[reservation.get_reservation_id()] = entry

    def add_to_list(self, *, entry: ReservationWrapper):
        """
        Adds the entry to the linked list. Maintains the list in sorted order.
        Cost: O(log(n)).
        @params entry : entry to add
        """
        bisect.insort_left(self.list, entry)

    def clear(self):
        """
        Clears the collection.
        """
        self.map.clear()
        self.list.clear()
        self.reservation_set.clear()

    def get_reservations(self, *, time: int = None, rtype: ResourceType = None) -> ReservationSet:
        """
        Performs an intersection query: returns all reservations from the
        specified resource type present in the collection that are active at the
        specified time instance.
        @params time : time instance
        @params rtype : resource type
        @returns reservations set containing active reservations
        """
        if time is None and rtype is None:
            return self.reservation_set

        result = ReservationSet()
        key = ReservationWrapper(reservation=None, start=time, end=time)

        # Find the location of key in the list.
        index = binary_search(a=self.list, x=key)
        if index < 0:
            index = -index -1

        # Scan the upper part of the list. We need to scan the whole list.
        i = index
        count = self.size()
        while i < count:
            entry = self.list[i]

            if rtype is None or rtype == entry.reservation.getType() and entry.start <= time <= entry.end:
                result.add(reservation=entry.reservation)
            i += 1

        # Scan the lower part of the list until no further intersections are possible
        i = index - 1
        while i >= 0:
            entry = self.list[i]
            if entry.end < time:
                break

            if entry.start <= time and rtype is None or entry.reservation.getType() == rtype:
                result.add(reservation=entry.reservation)
            i -= 1

        return result

    def remove_from_list(self, *, entry: ReservationWrapper):
        """
        Removes a reservation from the collection.
        @params reservation : reservation to remove
        """
        index = binary_search(a=self.list, x=entry)

        if index >= 0:
            self.list.pop(index)

    def remove_reservation(self, *, reservation: IReservation):
        """
        Removes a reservation from the collection.
        @params reservation : reservation to remove
        """
        if reservation.get_reservation_id() in self.map:
            entry = self.map[reservation.get_reservation_id()]
            self.map.pop(reservation.get_reservation_id())
            self.reservation_set.remove(reservation=reservation)
            self.remove_from_list(entry=entry)

    def size(self) -> int:
        """
        Returns the size of the collection.
        @returns size of the collection
        """
        return self.reservation_set.size()

    def tick(self, *, time: int):
        """
        Removes all reservations that have end time not after the given cycle.
        @params time : time
        """
        while True:
            if len(self.list) > 0:
                entry = self.list[0]
                if entry.end <= time:
                    self.list.remove(entry)
                    self.reservation_set.remove(reservation=entry.reservation)
                    self.map.pop(entry.reservation.get_reservation_id())
                else:
                    break
            else:
                break



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING
from fabric.actor.core.util.rpc_exception import RPCException
from fabric.actor.core.kernel.failed_rpc import FailedRPC
from fabric.actor.core.kernel.failed_rpc_event import FailedRPCEvent

if TYPE_CHECKING:
    from fabric.actor.core.kernel.rpc_request import RPCRequest


class RPCExecutor:
    """
    Execute an RPC
    """
    def __init__(self, *, request: RPCRequest):
        self.request = request
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()

    def post_exception(self, *, e: RPCException):
        """
        Handle any exception raised during processing
        """
        try:
            self.logger.error("An error occurred while performing RPC. Error type={} {}".format(e.get_error_type(), e))
            from fabric.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
            RPCManagerSingleton.get().remove_pending_request(self.request.request.get_message_id())
            failed = FailedRPC(e=e, request=self.request)
            self.request.actor.queue_event(FailedRPCEvent(actor=self.request.actor, failed=failed))
        except Exception as e:
            self.logger.error("postException failed = {}".format(e))

    def run(self):
        """
        Execute RPC
        """
        self.logger.debug("Performing RPC: type={} to:{}".format(self.request.request.get_type(),
                                                                 self.request.proxy.get_name()))
        try:
            self.request.proxy.execute(request=self.request.request)
        except RPCException as e:
            self.post_exception(e=e)
        finally:
            self.logger.debug("Completed RPC: type= {} to: {}".format(self.request.request.get_type(),
                                                                      self.request.proxy.get_name()))
            from fabric.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
            RPCManagerSingleton.get().de_queued()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_actor_runnable import IActorRunnable
from fabric.actor.core.apis.i_timer_queue import ITimerQueue
from fabric.actor.core.apis.i_actor_identity import IActorIdentity
from fabric.actor.core.apis.i_slice_operations import ISliceOperations
from fabric.actor.core.apis.i_reservation_operations import IReservationOperations
from fabric.actor.core.apis.i_tick import ITick

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor_event import IActorEvent
    from fabric.actor.core.apis.i_actor_proxy import IActorProxy
    from fabric.actor.core.apis.i_base_plugin import IBasePlugin
    from fabric.actor.core.apis.i_query_response_handler import IQueryResponseHandler
    from fabric.actor.core.kernel.failed_rpc import FailedRPC
    from fabric.actor.core.time.actor_clock import ActorClock
    from fabric.actor.core.util.id import ID
    from fabric.actor.security.auth_token import AuthToken
    from fabric.actor.core.apis.i_policy import IPolicy


class ActorType(Enum):
    """
    Enum for Actor Type
    """
    All = 0
    Orchestrator = 1
    Broker = 2
    Authority = 3

    @staticmethod
    def get_actor_type_from_string(*, actor_type: str) -> ActorType:
        """
        Convert String to Actor Type
        @param actor_type actor type
        """
        if actor_type.lower() == ActorType.Orchestrator.name.lower():
            return ActorType.Orchestrator
        if actor_type.lower() == ActorType.Broker.name.lower():
            return ActorType.Broker
        if actor_type.lower() == ActorType.Authority.name.lower():
            return ActorType.Authority
        return ActorType.All


class IActor(IActorIdentity, ISliceOperations, IReservationOperations, ITick, ITimerQueue):
    """
     IActor defines the common functionality of all actors. An actor
     offers a collection of management operations for slices and reservations and
     implements the public methods necessary to serve calls from other actors,
     e.g., requests for tickets and leases.

     Every actor has a globally unique identifier and a name. The current
     implementation assumes that names are globally unique. In addition, each
     actor can have an optional description (used for display purposes, e.g., the
     web portal).

     The actions of each actor, e.g, how to request new resources, how to
     arbitrate among multiple ticket requests, etc, are driven by policy modules.

     There are three types of actors:
         Orchestrator -
         Broker - arbiter among requests for resources. Brokers determine who gets what and for how long.
         Aggregate Manager - owner of resources.

         Each of the aforementioned roles is defined in a corresponding interface. An
         actor instance must implement at least one of these interfaces.
    """
    PropertyGuid = "ActorGuid"
    PropertyName = "ActorName"
    PropertyType = "ActorType"

    @abstractmethod
    def actor_added(self):
        """
        Informs the actor that it has been integrated in the container. This
        method should finish the initialization of the actor: some initialization
        steps may not be able to execute until the actor is part of the running
        container.

        Raises:
            Exception: if a critical error occurs while processing the event
        """

    @abstractmethod
    def actor_removed(self):
        """
        Informs the actor that it has been removed. This method should finish the
        shutdown/cleanup of the actor.
        """

    @abstractmethod
    def get_actor_clock(self) -> ActorClock:
        """
        Returns the actor clock used by the actor.

        Returns:
            actor clock
        """

    @abstractmethod
    def get_current_cycle(self) -> int:
        """
        Returns the cycle this actor is processing.

        Returns:
            current clock cycle
        """

    @abstractmethod
    def get_description(self) -> str:
        """
        Returns the description for the actor.

        Returns:
            description for the actor.
        """

    @abstractmethod
    def get_policy(self) -> IPolicy:
        """
        Returns the policy used by the actor.

        Returns:
            policy used by the actor.
        """

    @abstractmethod
    def get_plugin(self) -> IBasePlugin:
        """
        Returns the plugin used by the actor.

        Returns:
            plugin used by the actor.
        """

    @abstractmethod
    def get_type(self) -> ActorType:
        """
        Returns the type of the actor.

        Returns:
            type of the actor.
        """

    @abstractmethod
    def initialize(self):
        """
        Initializes the actor.

        Raises:
            Exception: if a critical error occurs while initialization
        """

    @abstractmethod
    def is_recovered(self) -> bool:
        """
        Checks if the actor has completed recovery.

        Returns:
            true if this actor has completed recovery
        """

    @abstractmethod
    def is_stopped(self) -> bool:
        """
        Checks if the actor has completed stopped.

        Returns:
            true if this actor has been stopped
        """

    @abstractmethod
    def recover(self):
        """
        Recovers the actor from saved state.

        Raises:
            Exception: if an error occurs during recovery
        """

    @abstractmethod
    def set_actor_clock(self, *, clock: ActorClock):
        """
        Sets the actor clock to be used by the actor.

        Args:
            clock: actor clock
        """

    @abstractmethod
    def set_description(self, *, description: str):
        """
        Sets the description for the actor.

        Args:
            description: actor description
        """

    @abstractmethod
    def set_identity(self, *, token: AuthToken):
        """
        Sets the identity of this actor. Must be called before initialize.

        Args:
            token: actor's identity token
        """

    @abstractmethod
    def set_policy(self, *, policy: IPolicy):
        """
        Sets the policy of this actor. Must be called before initialize.

        Args:
            policy: policy implementation to use
        """

    @abstractmethod
    def set_recovered(self, *, value: bool):
        """
        Sets the recovered flag.

        Args:
            value: flag value
        """

    @abstractmethod
    def set_plugin(self, *, plugin: IBasePlugin):
        """
        Sets the plugin of this actor. Must be called before initialize.

        Args:
            plugin: plugin to use.
        """

    @abstractmethod
    def start(self):
        """
        Performs all required actions when starting an actor.
        """

    @abstractmethod
    def stop(self):
        """
        Performs all required actions when stopping an actor.
        """

    @abstractmethod
    def queue_event(self, *, incoming: IActorEvent):
        """
        Adds an event.

        Args:
            incoming: incoming event
        """

    @abstractmethod
    def query(self, *, query: dict = None, caller: AuthToken = None,
              actor_proxy: IActorProxy = None, handler: IQueryResponseHandler = None,
              id_token: str = None):
        """
        Processes a query request from the specified caller.

        Args:
            query: query
            caller: caller
            actor_proxy: actor proxy
            handler: handler
            id_token: id_token

        Returns:
            query response
        """

    @abstractmethod
    def execute_on_actor_thread_and_wait(self, *, runnable: IActorRunnable):
        """
        Execute on Actor Thread and Wait until response is processed
        @params runnable: reservation to be processed
        """

    @abstractmethod
    def await_no_pending_reservations(self):
        """
        Await for pending reservations
        """

    @abstractmethod
    def get_logger(self):
        """
        Return the logger
        """

    @abstractmethod
    def handle_failed_rpc(self, *, rid: ID, rpc: FailedRPC):
        """
        Handle a failed rpc
        @params rid: reservation id
        @params rpc: failed rpc
        """

    @staticmethod
    def get_management_object_class() -> str:
        """
        Get Management Object class Name
        """

    @staticmethod
    def get_management_object_module() -> str:
        """
        Get Management Object class Module name
        """

    @staticmethod
    def get_kafka_service_class() -> str:
        """
        Get Kafka Service Class Name
        """

    @staticmethod
    def get_kafka_service_module() -> str:
        """
        Get Kafka Service Class Module Name
        """

    @staticmethod
    def get_mgmt_kafka_service_class() -> str:
        """
        Get Management Kafka Service Class Name
        """

    @staticmethod
    def get_mgmt_kafka_service_module() -> str:
        """
        Get Management Kafka Service Class Module Name
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import BrokerException
from fabric.actor.core.kernel.reservation_states import ReservationStates
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.prop_list import PropList
from fabric.actor.core.util.reservation_set import ReservationSet
from fabric.actor.core.policy.broker_priority_policy import BrokerPriorityPolicy
from fabric.actor.core.policy.inventory import Inventory
from fabric.actor.core.apis.i_client_reservation import IClientReservation

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker import IBroker
    from fabric.actor.core.policy.inventory_for_type import InventoryForType
    from fabric.actor.core.util.resource_type import ResourceType
    from fabric.actor.core.kernel.resource_set import ResourceSet


class BrokerSimplerUnitsPolicy(BrokerPriorityPolicy):
    def __init__(self, *, actor: IBroker = None):
        super().__init__(actor=actor)
        self.inventory = Inventory()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['actor']
        del state['clock']
        del state['initialized']

        del state['delegations']
        del state['combined_broker_model']
        del state['for_approval']
        del state['lock']

        del state['calendar']

        del state['last_allocation']
        del state['allocation_horizon']
        del state['ready']
        del state['queue']
        del state['inventory']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.actor = None
        self.clock = None
        self.initialized = False

        self.delegations = {}
        self.combined_broker_model = None
        self.load_combined_broker_model()

        self.lock = threading.Lock()
        self.calendar = None

        if self.required_approval:
            self.for_approval = ReservationSet()

        self.last_allocation = -1
        self.allocation_horizon = 0
        self.ready = False

        self.queue = None
        self.inventory = Inventory()

    def donate_reservation(self, *, reservation: IClientReservation):
        super().donate_reservation(reservation=reservation)
        self.inventory.get_new(reservation=reservation)

    def allocate(self, *, cycle: int):
        if self.get_next_allocation(cycle=cycle) != cycle:
            return

        self.last_allocation = cycle

        start_cycle = self.get_start_for_allocation(allocation_cycle=cycle)
        advance_cycle = self.get_end_for_allocation(allocation_cycle=cycle)
        requests = self.calendar.get_all_requests(cycle=advance_cycle)

        if requests is None or requests.size() == 0 and self.queue is None or self.queue.size() == 0:
            self.logger.debug("no requests for auction start cycle {}".format(start_cycle))
            return

        self.logger.debug("allocating resources for cycle {}".format(start_cycle))

        self.allocate_extending_reservation_set(requests=requests, start_cycle=start_cycle)
        self.allocate_queue(start_cycle=start_cycle)
        self.allocate_ticketing(requests=requests, start_cycle=start_cycle)

    def get_default_pool_id(self) -> ResourceType:
        result = None

        if len(self.inventory.get_inventory().keys()) > 0:
            result = self.inventory.get_inventory().keys().__iter__().__next__()

        return result

    def allocate_extending_reservation_set(self, *, requests: ReservationSet, start_cycle: int):
        if requests is not None:
            for reservation in requests.values():
                if reservation.is_extending_ticket() and not reservation.is_closed():
                    start = reservation.get_requested_term().get_new_start_time()
                    end = self.align_end(when=reservation.get_requested_term().get_end_time())

                    pool_id = self.get_current_pool_id(reservation=reservation)

                    inv = self.inventory.get(resource_type=pool_id)

                    if inv is not None:
                        ext_term = Term(start=reservation.get_term().get_start_time(), end=end, new_start=start)
                        self.extend_private(reservation=reservation, inv=inv, term=ext_term)
                    else:
                        reservation.fail(message=Constants.no_pool)

    def allocate_ticketing(self, *, requests: ReservationSet, start_cycle: int):
        if requests is not None:
            for reservation in requests.values():
                if not reservation.is_ticketing():
                    continue

                if self.ticket(reservation=reservation, start_cycle=start_cycle):
                    continue

                if self.queue is None and not reservation.is_failed():
                    reservation.fail(message="Insufficient resources")
                    continue

                if not reservation.is_failed():
                    if PropList.is_elastic_time(rset=reservation.get_requested_resources()):
                        self.logger.debug("Adding reservation + {} to the queue".format(reservation.get_reservation_id()))
                        self.queue.add(reservation=reservation)
                    else:
                        reservation.fail("Insufficient resources for specified start time, Failing reservation: {}".
                                         format(reservation.get_reservation_id()))

    def allocate_queue(self, *, start_cycle: int):
        if self.queue is None:
            return

        for reservation in self.queue.values():
            if not self.ticket(reservation=reservation, start_cycle=start_cycle):
                request_properties = reservation.get_requested_resources().get_request_properties()
                threshold = request_properties[BrokerPriorityPolicy.QueueThreshold]
                start = self.clock.cycle(when=reservation.get_requested_term().get_new_start_time())

                if threshold != 0 and ((start_cycle - start) > threshold):
                    reservation.fail_warn(message="Request has exceeded its threshold on the queue {}".format(
                        reservation))
                    self.queue.remove(reservation=reservation)
            else:
                self.queue.remove(reservation=reservation)

    def ticket(self, *, reservation: IBrokerReservation, start_cycle: int) -> bool:
        start_time = self.clock.date(cycle=start_cycle)

        self.logger.debug("cycle: {} new ticket request: {}".format(self.actor.get_current_cycle(), reservation))

        start = self.align_start(when=reservation.get_requested_term().get_new_start_time())
        end = self.align_end(when=reservation.get_requested_term().get_end_time())

        term = None
        if start < start_time and PropList.is_elastic_time(rset=reservation.get_requested_resources()):
            length = ActorClock.to_milliseconds(when=end) - ActorClock.to_milliseconds(when=start)

            start = self.clock.cycle_start_in_millis(cycle=start_cycle)
            term = Term(start=start, length=length)

        pool_id = reservation.get_requested_resources().get_type()
        if pool_id is None or not self.inventory.contains_type(resource_type=pool_id):
            pool_id = self.get_default_pool_id()

        if pool_id is not None:
            inv = self.inventory.get(resource_type=pool_id)

            if inv is not None:
                if term is None:
                    term = Term(start=start, end=end)

                return self.ticket_inventory(reservation=reservation, inv=inv, term=term)
            else:
                reservation.fail(message=Constants.no_pool)
        else:
            reservation.fail(message=Constants.no_pool)

        return False

    def ticket_inventory(self, *, reservation: IBrokerReservation, inv: InventoryForType, term: Term) -> bool:
        try:
            rset = reservation.get_requested_resources()
            needed = rset.get_units()
            available = inv.get_free()
            to_allocate = min(needed, available)

            if to_allocate == 0:
                return False

            if to_allocate < needed and not PropList.is_elastic_size(rset=reservation.get_requested_resources()):
                return False

            properties = inv.allocate(count=to_allocate, request=rset.get_request_properties())
            properties = PropList.merge_properties(incoming=inv.get_properties(), outgoing=properties)

            if to_allocate < needed:
                self.logger.error("partially satisfied request: allocated= {} needed={}".format(to_allocate, needed))

            self.issue_ticket(reservation=reservation, units=to_allocate, rtype=inv.get_type(), term=term,
                              properties=properties, source=inv.get_source())
            return True
        except Exception as e:
            self.logger.error(e)
            reservation.fail(message=str(e))
            return False

    def extend_private(self, *, reservation: IBrokerReservation, inv: InventoryForType, term: Term):
        try:
            rset = reservation.get_requested_resources()
            needed = rset.get_units()
            current = reservation.get_resources().get_units()
            difference = needed - current

            units = current

            properties = None

            if difference > 0:
                available = inv.get_free()
                to_allocate = min(difference, available)

                if to_allocate > 0:
                    properties = inv.allocate(count=to_allocate, request=rset.get_request_properties(),
                                              resource=rset.get_resource_properties())

                if to_allocate < difference:
                    self.logger.error("partially satisfied request: allocated= {} needed={}".format(
                        to_allocate, difference))

                units += to_allocate
            elif difference < 0:
                properties = inv.free(count=-difference, request=rset.get_request_properties(),
                                      resource=rset.get_resource_properties())
                units += difference

            properties = PropList.merge_properties(incoming=inv.get_properties(), outgoing=properties)

            self.issue_ticket(reservation=reservation, units=units, rtype=inv.get_type(), term=term,
                              properties=properties, source=inv.get_source())
        except Exception as e:
            self.logger.error(e)
            reservation.fail(message=str(e), exception=e)

    def issue_ticket(self, *, reservation: IBrokerReservation, units: int, rtype: ResourceType,
                     term: Term, properties: dict, source: IClientReservation):

        delegation = self.actor.get_plugin().get_ticket_factory().make_delegation(
            units=units, term=term, rtype=rtype, properties=properties,
            holder=self.get_client_id(reservation=reservation))

        mine = self.extract(source=source.get_resources(), delegation=delegation)

        p = reservation.get_requested_resources().get_request_properties()
        mine.set_request_properties(p=p)

        if mine is not None and not reservation.is_failed():
            reservation.set_approved(term=term, approved_resources=mine)
            reservation.set_source(source=source)
            self.logger.debug("allocated: {} for term: {}".format(mine.get_units(), term))
            self.logger.debug("resourceshare= {} mine= {}".format(units, mine.get_units()))

            if self.required_approval:
                self.add_for_approval(reservation=reservation)
            else:
                self.add_to_calendar(reservation=reservation)
                reservation.set_bid_pending(value=False)
        else:
            if mine is None:
                raise BrokerException("There was an error extracting a ticket from the source ticket")

    def release(self, *, reservation):
        if isinstance(reservation, IBrokerReservation):
            self.logger.debug("Broker reservation")
            super().release(reservation=reservation)
            if reservation.is_closed_in_priming():
                self.logger.debug("Releasing resources (closed in priming)")
                self.release_resources(rset=reservation.get_approved_resources(),
                                       term=reservation.get_approved_term())
            else:
                self.logger.debug("Releasing resources")
                self.release_resources(rset=reservation.get_resources(),
                                       term=reservation.get_term())
        elif isinstance(reservation, IClientReservation):
            self.logger.debug("Client reservation")
            super().release(reservation=reservation)
            status = self.inventory.remove(source=reservation)
            self.logger.debug("Removing reservation: {} from inventory status: {}".format(
                reservation.get_reservation_id(), status))

    def release_not_approved(self, *, reservation: IBrokerReservation):
        super().release_not_approved(reservation=reservation)
        self.release_resources(rset=reservation.get_approved_resources(), term=reservation.get_approved_term())

    def release_resources(self, *, rset: ResourceSet, term: Term):
        try:
            if rset is None or term is None or rset.get_resources() is None:
                self.logger.warning("Reservation does not have resources to release")
                return
            inv = self.inventory.get(resource_type=rset.get_type())
            if inv is None:
                raise BrokerException("Cannot release resources: missing inventory")
            inv.free(count=rset.get_units(), resource=rset.get_resource_properties())
        except Exception as e:
            self.logger.error("release resources {}".format(e))

    def revisit_server(self, *, reservation: IBrokerReservation):
        super().revisit_server(reservation=reservation)

        if reservation.get_state() == ReservationStates.Ticketed:
            self.revisit_ticketed(reservation=reservation)

    def revisit_ticketed(self, *, reservation: IBrokerReservation):
        rset = reservation.get_resources()
        rtype = rset.get_type()
        inv = self.inventory.get(resource_type=rtype)
        if inv is None:
            raise BrokerException("cannot free resources: no inventory")

        inv.allocate_revisit(count=rset.get_units(), resource=rset.get_resource_properties())


